<?php

namespace Picea\Ui\Method;

use Psr\Http\Message\ServerRequestInterface,
    Psr\Http\Message\ResponseInterface;

class FormContext implements FormContextInterface
{
    public string $formContextMethod;

    public string $formName;

    public bool $formSent = false;

    public bool $formExecuted = false;

    public mixed $formExecutionStatus = null;

    public array $formContextValues = [];

    public array $formContextFiles = [];

    public array $formContextMessages = [];

    public bool $formSkipCsrf = false;

    protected array $formContextCatchedMethods = [ 'POST', 'PUT', 'PATCH', 'DELETE', ];

    public ServerRequestInterface $formContextRequest;
    
    public ? ResponseInterface $formContextResponse = null;

    public FormInterface $form;

    public function __construct(ServerRequestInterface $request, ? string $formName = null)
    {
        $this->formContextRequest = $request;

        if ( $formName ) {
            $this->formName = $formName;
        }

        $this->formContextValues = $request->getParsedBody() ?: [];

        if ( ! $this->formContextValues ) {
            $content = mb_convert_encoding((string) $request->getBody(), 'UTF-8');

            if ( $content && ( $json = json_decode($content, true) ) ) {
                $this->formContextValues = $json;
            }
        }

        $this->fillValues();

        $this->formContextFiles = $request->getUploadedFiles() ?: [];

        $this->initialize();
    }

    public function initialize() : void {}
    
    public function valid() : bool
    {
        foreach($this->formContextMessages as $message) {
            if ( $message->isError() ) {
                return false;
            }
        }
        
        return true;
    }

    public function executed() : bool
    {
        return $this->formExecutionStatus === false ? false : $this->formExecuted;
    }

    public function formSent() : bool
    {
        $valid = in_array($this->requestMethod(), $this->formContextCatchedMethods);

        if ( ! $this->formSkipCsrf && ($this->formName ?? false) ) {
            $token = $this->get('picea-ui-form')[$this->formName] ?? false;

            if ( $token ) {
                if ($this->validateCsrfToken) {
                    $valid = in_array($token, $_SESSION["picea-ui.form:{$this->formName}"] ?? []);
                }
                else {
                    $valid = (bool) $token;
                }
            }
            else {
                $valid = false;
            }
        }

        return $this->formSent = $valid;
    }

    public function getExecutionStatus() : mixed
    {
        return $this->formExecutionStatus;
    }

    public function getRequest() : ServerRequestInterface
    {
        return $this->formContextRequest;
    }

    public function requestMethod() : string
    {
        return $this->formContextRequest->getMethod();
    }

    public function __set($key, $value)
    {
        return $this->set($key, $value);
    }

    public function __get($key)
    {
        return $this->get($key);
    }

    public function __isset($key)
    {
        return array_key_exists($key, $this->formContextValues);
    }

    public function __unset($key)
    {
        $this->delete($key);
    }   
    
    public function get(string $key, $default = null)
    {
        return $this->has($key) ? $this->formContextValues[$key] : $default;
    }

    public function set(string $key, $value)
    {
        if ($this->canWriteProperty($key)) {
            $this->$key = $value;
        }

        return $this->formContextValues[$key] = $value;
    }

    public function sets(array $values) : void
    {
        foreach($values as $key =>$value) {
            $this->set($key, $value);
        }
    }


    public function delete(string $key) : void
    {
        unset($this->formContextValues[$key]);
    }

    public function has(string $key) : bool
    {
        return array_key_exists($key, $this->formContextValues);
    }
    
    public function pushMessage(FormMessage $message) : void
    {
        $this->formContextMessages[] = $message;
    }

    protected function fillValues() : void
    {

        foreach($this->formContextValues as $property => $value) {
            if ($this->canWriteProperty($property)) {
                $this->$property = $value;
            }
        }
    }

    protected function canWriteProperty(string $property) : bool
    {
        # Skipping overrides of this particular class vars as a security measure
        static $skipping = array_keys(array_change_key_case(get_class_vars(FormContext::class), CASE_LOWER));

        return ! in_array(strtolower($property), $skipping) && property_exists($this, $property);
    }
}
